#!/bin/bash
#
#  build-initrd.sh
#
#  Written by Jari Ruusu, December 12 2001
#
#  Copyright 2001 by Jari Ruusu.
#  Redistribution of this file is permitted under the GNU Public License.
#
#  Changes by Hauke Johannknecht <ash@ash.de> 11/2001
#     - added Pivot mode
#     - added conffile-loading
#     - added initrdonly-mode
#     - added device 0 nodes
#
#
#  kernel .config :  CONFIG_BLK_DEV_RAM=y
#                    CONFIG_BLK_DEV_RAM_SIZE=4096
#                    CONFIG_BLK_DEV_INITRD=y
#                    CONFIG_MINIX_FS=y
#                    CONFIG_PROC_FS=y
#
#                    USEPIVOT=0                 USEPIVOT=1
#                    ----------                 ----------
#  /etc/lilo.conf :  initrd=/boot/initrd.gz     append="init=/linuxrc"
#                    root=/dev/ram1             initrd=/boot/initrd.gz
#                                 ^             root=/dev/ram0
#                                                            ^
#  usage :  ./build-initrd.sh [configfile]
#           lilo
#           cp -p /lib/modules/`uname -r`/block/loop.o /boot/loop-`uname -r`.o
#

### All default-values can be altered via the configfile

# normal /boot partition
BOOTDEV=/dev/hda1

# /boot partition type
BOOTTYPE=ext2

# encrypted root partition
CRYPTROOT=/dev/hda2

# root partition type
ROOTTYPE=ext2

# encryption type (AES128 / AES192 / AES256) of root partition
CIPHERTYPE=AES128

# optional password seed for root partition
#PSEED="-S XXXXXX"

# 1 = use loop module, 0 = loop driver linked to kernel
USEMODULE=1

# 1 = use pivot_root, 0 = use old change_root
# See above header for root= and append="init=" lilo.conf definitions.
# pivot_root is not available on 2.2 and 2.0 kernels.
USEPIVOT=0

# 1 = stop after creating and copying initrd, 0 = also copy tools/libs
INITRDONLY=0

# source root directory where files are copied from
SOURCEROOT=

# destination root directory where files are written to
DESTINATIONROOT=

# dest-dir below dest-root
DESTINATIONPREFIX=/boot

# encrypted root loop device index (0 ... 7), 5 == /dev/loop5
# _must_ match /etc/fstab entry:   /dev/loop5  /  ext2  defaults,xxxx  0  1
ROOTLOOPINDEX=5

# temporary loop device index used in this script, 7 == /dev/loop7
TEMPLOOPINDEX=7


if [ $# = 1 ] ; then
    if [ ! -f $1 ] ; then
        echo "ERROR: Can't find configfile '$1'"
        echo "Usage: $0 [configfile]"
        exit 1
    fi
    echo "Loading config from '$1'"
    . $1
fi

set -e
umask 077
cat - <<EOF >tmp-c-$$.c

/* Note: this program does not initialize C library, so all level 3     */
/* library calls are forbidden. Only level 2 systems calls are allowed. */

#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/mount.h>
#include <sys/utsname.h>
#include <sys/stat.h>
#include <fcntl.h>

#if ${USEPIVOT}
# ifdef __ia64__
#  include <sys/syscall.h>
#  define pivot_root(new_root,put_old) syscall(SYS_pivot_root,new_root,put_old)
# else
#  include <linux/unistd.h>
#  if !defined(__NR_pivot_root) && defined(__i386__)
#   define __NR_pivot_root 217
#  endif
static _syscall2(int,pivot_root,const char *,new_root,const char *,put_old)
# endif
static char * argv_init[] = { "init", 0, };
static char * envp_init[] = { "HOME=/", "TERM=linux", 0, };
#endif

void strCat(char *d, char *s)
{
    while(*d) d++;
    while(*s) *d++ = *s++;
    *d = 0;
}

void wrStr(char *s)
{
    char *p = s;
    int x = 0;

    while(*p) p++, x++;
    write(1, s, x);
}

int exeWait(char *p)
{
    int x, y;
    char *a[50], *e[1];

    if(!(x = fork())) {
        while(*p && (x < ((sizeof(a) / sizeof(char *)) - 1))) {  
            a[x++] = p;
            while(*p && (*p != ' ') && (*p != '\t')) p++;
            while((*p == ' ') || (*p == '\t')) *p++ = 0;
        }
        e[0] = a[x] = 0;
        if(x) execve(a[0], &a[0], &e[0]);
        _exit(1); 
    }
    waitpid(x, &y, 0);
    if(!WIFEXITED(y) || (WEXITSTATUS(y) != 0)) {
        wrStr("Command \""); wrStr(p); wrStr("\" returned error\n");
        return(1);
    }
    return(0);
}

void _start()
{
#if !(${USEPIVOT})
    int x;
#endif
#if ${USEMODULE}
    struct utsname un;
#endif
    char buf[300];

    /* this intentionally mounts /boot partition as /lib */
    if(mount("${BOOTDEV}", "/lib", "${BOOTTYPE}", MS_MGC_VAL | MS_RDONLY, 0)) {
        wrStr("Mounting ${BOOTDEV} as /lib failed\n");
        goto fail4;
    }

#if ${USEMODULE}
    uname(&un);
    buf[0] = 0;
    strCat(buf, "/lib/insmod -o loop /lib/loop-");
    strCat(buf, &un.release[0]);
    strCat(buf, ".o");
    if(exeWait(buf)) goto fail3;
#endif

    wrStr("\nEncrypted file system, please supply correct password to continue\n\n");
    buf[0] = 0;
    strCat(buf, "/lib/losetup -e ${CIPHERTYPE} ${PSEED} /dev/loop${ROOTLOOPINDEX} ${CRYPTROOT}");
    if(exeWait(buf)) goto fail3;
    wrStr("\n");

#if ${USEPIVOT}
    umount("/lib");

    if(mount("/dev/loop${ROOTLOOPINDEX}", "/new-root", "${ROOTTYPE}", MS_MGC_VAL | MS_RDONLY, 0)) {
        wrStr("Looks like you didn't say the magic word. Mounting /dev/loop${ROOTLOOPINDEX} failed\n");
        goto fail4;
    }
    if(chdir("/new-root")) {
        wrStr("chdir() to /new-root failed\n");
        goto fail1;
    }
    if(pivot_root(".", "initrd")) {
        wrStr("pivot_root() to new root failed. Older kernels don't have pivot_root()\n");
        fail1:
        chdir("/");
        umount("/new-root");
        goto fail4;
    }

    close(0);
    close(1);
    close(2);
    open("dev/console", O_RDWR, 0);
    dup(0);
    dup(0);

    if(chroot(".")) {
        wrStr("chroot() to new root failed\n");
        goto fail1;
    }
    wrStr("Pivoting to encrypted root completed successfully\n");

    execve("/sbin/init", argv_init, envp_init);
    execve("/etc/init", argv_init, envp_init);
    execve("/bin/init", argv_init, envp_init);
    wrStr("Bummer! Launching init failed\n");
    goto fail4;
#else
    if(mount("none", "/proc", "proc", MS_MGC_VAL, 0)) {
        wrStr("Mounting /proc failed\n");
        goto fail3;
    }
    if((x = open("/proc/sys/kernel/real-root-dev", O_WRONLY, 0)) == -1) {
        wrStr("Unable to open real-root-dev\n");
        goto fail2;
    }
    write(x, "0x70${ROOTLOOPINDEX}\n", 6);
    close(x);
    fail2:
    umount("/proc");
#endif

    fail3:
    umount("/lib");
    fail4:
    _exit(0);
}
EOF

gcc -Wall -O2 -s -static -nostartfiles -pipe tmp-c-$$.c -o tmp-c-$$
rm -f tmp-c-$$.[co]

x=`cat tmp-c-$$ | wc -c`
y=`expr ${x} + 1023`
x=`expr ${y} / 1024`
y=`expr ${x} + 12`

dd if=/dev/zero of=tmp-i-$$ bs=1024 count=${y}
losetup /dev/loop${TEMPLOOPINDEX} tmp-i-$$
mkfs -t minix -i 32 /dev/loop${TEMPLOOPINDEX}
losetup -d /dev/loop${TEMPLOOPINDEX}
mkdir tmp-d-$$
mount -t minix tmp-i-$$ tmp-d-$$ -o loop=/dev/loop${TEMPLOOPINDEX}
cd tmp-d-$$

mkdir dev lib
mv ../tmp-c-$$ linuxrc

if [ ${USEPIVOT} == 1 ] ; then
    mkdir new-root
else
    mkdir proc
fi

mknod dev/console c 5 1
mknod dev/tty c 5 0
mknod dev/tty1 c 4 1
mknod dev/null c 1 3
mknod dev/zero c 1 5
mknod dev/ram1 b 1 1
mknod dev/loop${ROOTLOOPINDEX} b 7 ${ROOTLOOPINDEX}

# <device name prefix> <major dev-id> <minor dev-id start> <device 0 suffix>
function maybeMakeDiskNode
{
    x=$3
    for i in "$4" 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15 ; do
        for y in ${BOOTDEV} ${CRYPTROOT} ; do
            if [ ${y} == /dev/$1${i} ] ; then
                mknod dev/$1${i} b $2 ${x}
            fi
        done
        x=`expr ${x} + 1`
    done
    return
}
maybeMakeDiskNode hda  3  0 ""
maybeMakeDiskNode hdb  3 64 ""
maybeMakeDiskNode hdc 22  0 ""
maybeMakeDiskNode hdd 22 64 ""
maybeMakeDiskNode sda  8  0 ""
maybeMakeDiskNode sdb  8 16 ""
maybeMakeDiskNode sdc  8 32 ""
maybeMakeDiskNode sdd  8 48 ""
maybeMakeDiskNode scd 11  0  0

cd ..
df tmp-d-$$
umount tmp-d-$$
rmdir tmp-d-$$
sync ; sync ; sync
gzip -9 tmp-i-$$
mv tmp-i-$$.gz ${DESTINATIONROOT}${DESTINATIONPREFIX}/initrd.gz
ls -l ${DESTINATIONROOT}${DESTINATIONPREFIX}/initrd.gz

if [ ${INITRDONLY} == 1 ] ; then
    echo Done.
    sync
    exit 0
fi

z="/sbin/losetup"
if [ ${USEMODULE} == 1 ] ; then
    z="/sbin/insmod /sbin/losetup"
fi
for x in ${z} ; do
    echo Copying ${SOURCEROOT}${x} to ${DESTINATIONROOT}${DESTINATIONPREFIX}
    cp -p ${SOURCEROOT}${x} ${DESTINATIONROOT}${DESTINATIONPREFIX}
    y=`ldd ${SOURCEROOT}${x} | perl -ne 'if(/ => ([^ ]*) /){print "$1\n"}'`
    for a in ${y} ; do
        echo Copying ${SOURCEROOT}${a} to ${DESTINATIONROOT}${DESTINATIONPREFIX}
        cp -p ${SOURCEROOT}${a} ${DESTINATIONROOT}${DESTINATIONPREFIX}
    done
done

if [ ! -d ${DESTINATIONROOT}/initrd ] ; then
    mkdir ${DESTINATIONROOT}/initrd
fi

echo Done.
sync
exit 0
